from r3frame.globals import pg
from r3frame.utils import damp_lin

from r3frame.objects.game.animation import Animation

class Game_Object:
    def __init__(self, size: list[int]=[32, 32], color: list[int]=[0, 255, 0], location: list[int]=[0, 0], mass: float=100.0, speed: float=100.0, vthreshold: float=5.0):
        self.size = size
        self.color = color
        self.image = pg.Surface(size)
        self.image.fill(color)
        
        self.mass = mass
        self.speed = speed
        self.vthreshold = vthreshold

        self.location = location
        self.velocity = [0.0, 0.0]
        self.last_location = location
        
        self.action = None
        self.animations = {}
        self.animation = None

    def set_animation(self, action: str, frames: list[pg.Surface], loop: bool=True, frame_duration: float=5.0, frame_offset: list[int]=[0, 0]) -> None:
        try:
            self.animations[action] = Animation(frames, loop, frame_duration, frame_offset)
            self.animation = self.animations[action]
            self.action = action
        except Exception as err: pass

    def rem_animation(self, action: str) -> None:
        try:
            del self.animations[action]
        except KeyError as err: return None

    def set_action(self, action: str) -> None:
        if self.action == action: return None
        try:
            self.action = action
            self.animation.reset()
            self.animation = self.animations.get(action, None)
        except KeyError as err: return None

    def get_rect(self) -> pg.Rect: return pg.Rect(self.location, self.size)

    def set_image(self, image: pg.Surface) -> None:
        if isinstance(image, pg.Surface):
            self.image = image
        elif image is None:
            self.image = pg.Surface(self.size)
            self.image.fill(self.color)

    def set_velocity(self, vx: float=0.0, vy: float=0.0) -> None:
        if vx: self.velocity[0] = vx
        if vy: self.velocity[1] = vy

    def get_facing(self) -> str:
        dx = None
        dy = None
        v = self.velocity
        if v[0] < 0:
            dx = "L"
        if v[0] > 0:
            dx = "R"
        if v[1] < 0:
            dy = "U"
        if v[1] > 0:
            dy = "D"

        if dx and dy: return dx+dy
        return dx if dy is None else dy

    def update(self, delta_time: float) -> None:
        # Apply movement
        self.last_location = self.location
        self.location[0] += self.velocity[0] * delta_time
        self.location[1] += self.velocity[1] * delta_time
        self.velocity = [*map(lambda v: damp_lin(v, self.mass, self.vthreshold, delta_time), self.velocity)]

        if self.animation:
            self.animation.update()
            self.image = self.animation.get_frame()



from r3frame.utils import math
from r3frame.globals import pg
from r3frame.objects.game import Game_Object

class Grid_Map:
    def __init__(self, width: int, height: int, cell_size: int) -> None:
        self.cells = {}
        self.width = width                                  # in cells
        self.height = height                                # in cells
        self.cell_size = cell_size                          # in pixels
        self.size = [width * cell_size, height * cell_size] # in pixels

    def get_cell(self, x: int, y: int) -> Game_Object | None:
        """Returns the object at the given world (pixel) position."""
        if x < 0 or x >= self.size[0] or y < 0 or y >= self.size[1]: return
        gx, gy = x // self.cell_size, y // self.cell_size
        return self.cells.get((gx, gy), None)

    def set_cell(self, x: int, y: int, obj: Game_Object) -> None:
        """Places an object in the grid at the given world (pixel) position."""
        if x < 0 or x >= self.size[0] or y < 0 or y >= self.size[1]: return
        gx, gy = x // self.cell_size, y // self.cell_size
        obj.location = [gx * self.cell_size, gy * self.cell_size]
        self.cells[(gx, gy)] = obj

    def get_region(self, x: int, y: int) -> dict[str, Game_Object | None]:
        """Gets the object in the given cell and its immediate neighbors."""
        if x < 0 or x >= self.size[0] or y < 0 or y >= self.size[1]: return
        gx, gy = x // self.cell_size, y // self.cell_size
        directions = {
            "center": (0, 0),
            "up": (0, -1),
            "down": (0, 1),
            "left": (-1, 0),
            "right": (1, 0),
            "left-up": (-1, -1),
            "right-up": (1, -1),
            "left-down": (-1, 1),
            "right-down": (1, 1),
        }
        return {dir_name: self.cells.get((gx + dx, gy + dy)) for dir_name, (dx, dy) in directions.items()}
    
    def query_region(self, x: int, y: int, radius: int = 1) -> dict[tuple[int, int], Game_Object | None]:
        """Returns objects in a region centered at (x, y) within the given radius.
        
        Arguments:
        - x, y: The world (pixel) position to search around.
        - radius: The search radius in number of cells.
        
        Returns:
        - A dictionary where keys are grid coordinates (gx, gy) and values are the objects in those cells.
        """
        if x < 0 or x >= self.size[0] or y < 0 or y >= self.size[1]: 
            return {}

        gx, gy = x // self.cell_size, y // self.cell_size
        region = {}

        for dx in range(-radius, radius + 1):
            for dy in range(-radius, radius + 1):
                cell_x, cell_y = gx + dx, gy + dy
                if 0 <= cell_x < self.width and 0 <= cell_y < self.height:
                    region[(cell_x, cell_y)] = self.cells.get((cell_x, cell_y))

        return region

    def render_debug(self, renderer, center_x: int=0, center_y: int=0, radius: int=8):
        """Renders a portion of the grid around a center position within a given radius."""
        cell_w, cell_h = self.cell_size, self.cell_size
        radius_px = radius * cell_w  # Convert radius from cells to pixels

        # Convert world position to grid cell indices
        start_x = max((center_x - radius_px) // cell_w, 0)
        start_y = max((center_y - radius_px) // cell_h, 0)
        end_x = min((center_x + radius_px) // cell_w, self.width - 1)
        end_y = min((center_y + radius_px) // cell_h, self.height - 1)

        # Draw grid lines within the region
        for gx in range(start_x, end_x + 1):
            x = gx * cell_w
            renderer.draw_line([x, start_y * cell_h], [x, (end_y + 1) * cell_h], [125, 125, 125], 1)
        
        for gy in range(start_y, end_y + 1):
            y = gy * cell_h
            renderer.draw_line([start_x * cell_w, y], [(end_x + 1) * cell_w, y], [125, 125, 125], 1)

        # Draw objects within the visible region
        for (gx, gy), obj in self.cells.items():
            if start_x <= gx <= end_x and start_y <= gy <= end_y:
                obj_x, obj_y = gx * cell_w, gy * cell_h
                renderer.draw_circle([obj_x + self.cell_size / 2, obj_y + self.cell_size / 2], self.cell_size, [255, 255, 255], 1)



import r3frame as frame
import random

play = frame.app.Application("Playground-Fixed", "Playground-Fixed", "GRID")

def load_assets() -> None:
    play.assets.load_image_sheet("run", frame.utils._asset_path("images/run.png"), [32, 32])
    play.assets.load_image_sheet("idle", frame.utils._asset_path("images/idle.png"), [32, 32])
    
    play.add_button("button_1", frame.utils._asset_path("fonts/megamax.ttf"), "Hello!", [200, 50], location=[5, 550], padding=[65, 16], text_size=16)
    button_1 = play.get_button("button_1")
    
    def on_click() -> None:
        button_1.border_color = button_1.text_color = [
            random.randint(1, 255),
            random.randint(1, 255),
            random.randint(1, 255)
        ]
    def on_hover() -> None:
        button_1.text = "Click Me!"
        button_1.padding = [50, 16]
        button_1.border_color = button_1.text_color = [0, 255, 0]
    def on_unhover() -> None:
        button_1.text = "Hello!"
        button_1.padding = [65, 16]
        button_1.border_color = button_1.text_color = [255, 255, 255]
    button_1.on_click = on_click
    button_1.on_hover = on_hover
    button_1.on_unhover = on_unhover
play.load_assets = load_assets

def load_objects() -> None:
    play.player = frame.objects.game.Game_Object(
        mass=500, speed=150,
        location=[100, 100], color=[0, 0, 255],
    )

    play.player.set_animation("idle", play.assets.get_image("idle"), frame_duration=4)
    play.player.set_animation("run", play.assets.get_image("run"), frame_duration=5)

    def post_render():
        play.partition.render_debug(play.renderer, *[*map(int, [play.player.location[0] + play.player.size[0] / 2, play.player.location[1] + play.player.size[1] / 2])])
    play.renderer.post_render = post_render
play.load_objects = load_objects

def handle_events() -> None:
    zoom_factor = 3

    if play.events.key_pressed(frame.app.inputs.Keyboard.Escape): play.events.quit = 1

    if play.events.key_pressed(frame.app.inputs.Keyboard.F1): play.renderer.set_flag(play.renderer.FLAGS.SHOW_CAMERA)
    if play.events.key_pressed(frame.app.inputs.Keyboard.F2): play.renderer.rem_flag(play.renderer.FLAGS.SHOW_CAMERA)

    if play.events.key_held(frame.app.inputs.Keyboard.Down):    play.camera.set_velocity(vy=play.player.speed)
    if play.events.key_held(frame.app.inputs.Keyboard.Right):   play.camera.set_velocity(vx=play.player.speed)
    if play.events.key_held(frame.app.inputs.Keyboard.Left):    play.camera.set_velocity(vx=-play.player.speed)
    if play.events.key_held(frame.app.inputs.Keyboard.Up):      play.camera.set_velocity(vy=-play.player.speed)

    if play.events.mouse_wheel_up:
        play.camera.mod_viewport(-zoom_factor)

    if play.events.mouse_wheel_down:
        play.camera.mod_viewport(zoom_factor)

    if not isinstance(frame.app.inputs.Mouse.Hovering, frame.app.ui.Button) and play.events.mouse_held(frame.app.inputs.Mouse.LeftClick):
        mouse_location = frame.app.inputs.Mouse.get_location()
        world_x = (mouse_location[0] / play.camera.viewport_scale[0] + play.camera.location[0])
        world_y = (mouse_location[1] / play.camera.viewport_scale[1] + play.camera.location[1])
        play.partition.set_cell(world_x, world_y, frame.objects.game.Game_Object(
            size=[play.partition.cell_size, play.partition.cell_size],
            color=[random.randint(1, 255), random.randint(1, 255), random.randint(1, 255)]
        ))

    if not isinstance(frame.app.inputs.Mouse.Hovering, frame.app.ui.Button) and play.events.mouse_pressed(frame.app.inputs.Mouse.RightClick):
        mouse_location = frame.app.inputs.Mouse.get_location()
        world_x = (mouse_location[0] / play.camera.viewport_scale[0] + play.camera.location[0])
        world_y = (mouse_location[1] / play.camera.viewport_scale[1] + play.camera.location[1])
        for o in play.partition.query_region(world_x, world_y, 1).values():
            if o:
                o.color = [255, 255, 255]
                o.set_image(None)

    if play.events.key_held(frame.app.inputs.Keyboard.W):       play.player.set_velocity(vy=-play.player.speed)
    if play.events.key_held(frame.app.inputs.Keyboard.S):       play.player.set_velocity(vy=play.player.speed)
    if play.events.key_held(frame.app.inputs.Keyboard.D):
        play.player.animation.flip_x =False
        play.player.set_velocity(vx=play.player.speed)
    if play.events.key_held(frame.app.inputs.Keyboard.A):
        play.player.animation.flip_x = True
        play.player.set_velocity(vx=-play.player.speed)
play.handle_events = handle_events

def handle_update() -> None:
    if play.player.velocity[0] != 0 or play.player.velocity[1] != 0:
        play.player.set_action("run")
    else:
        play.player.set_action("idle")

    play.player.update(play.clock.delta)
    for tile in play.partition.cells:
        if hasattr(tile, "update"): tile.update(play.clock.delta)

    play.camera.center_on(play.player.size, play.player.location)
play.handle_update = handle_update

def handle_render() -> None:
    [play.renderer.draw_call(o.image, o.location) for o in [*play.partition.cells.values(), play.player] if hasattr(o, "image")]
    
    play.dev.set_text_field("FPS", f"{play.clock.FPS:0.1f}", color=[0, 255, 0] if play.clock.FPS > play.clock.maxFPS/2 else [255, 0, 0])
    play.dev.set_text_field("OBJECTS", f"{len([*play.partition.cells, play.player])}")
    play.dev.set_text_field("DIRECTION", f"{play.player.get_facing()}")
    play.dev.set_text_field("VELOCITY", f"{play.player.velocity[0]:0.1f}, {play.player.velocity[1]:0.1f}")
    play.dev.set_text_field("LOCATION", f"{play.player.location[0]:0.1f}, {play.player.location[1]:0.1f}")
play.handle_render = handle_render

def playground(): play.run()



import r3frame.utils as utils
import r3frame.version as version
import r3frame.application.ui as ui
import r3frame.application.inputs as inputs
import r3frame.application.events as events
import r3frame.application.resource as resource
import r3frame.application.resource as resource

from r3frame.objects.world import Grid_Map, Quad_Map

class Application:
    def __init__(self, name: str="My App", title: str="My Game", partition: str="GRID") -> None:
        self.clock = resource.Clock()
        self.events = events.Event_Manager()
        self.assets = resource.Asset_Manager()

        match partition.upper():
            case "GRID": self.partition = Grid_Map(1500, 1500, 8)
            case "QUAD": self.partition = Quad_Map(1500, 1500, 8)

        self.buttons = {}
        self.window = resource.Window([800, 600], self.partition.size)
    
        self.camera = resource.Camera(self.window)
        self.renderer = resource.Renderer(self.camera)

        self.dev = resource.DevDisplay(
            f"r3frame {version.R3FRAME_YEAR}.{version.R3FRAME_MINOR}.{version.R3FRAME_PATCH}",
            self.window, [100, 100], [0, 0], utils._asset_path("fonts/megamax.ttf"), text_size=20)

    def add_button(
            self, key: str, font_path:str, text: str="Button",
            size: list[int]=[64, 64], location: list[int|float]=[0, 0],
            color: list[int]=[0, 0, 0], text_color:list[int]=[255, 255, 255], text_size: int=18, padding: list[int]=[0, 0],
            border_size: list[int]=[5, 5], border_radius: list[int]=[0, 0, 0, 0], border_color: list[int]=[255, 255, 255],
        ) -> None:
        self.buttons[key] = ui.Button(
            font_path, text, size, location,
            color, text_color, text_size, padding,
            border_size, border_radius, border_color
        )
    
    def get_button(self, key: str) -> ui.Button|None:
        return self.buttons.get(key, None)

    def load_assets(self) -> None: raise NotImplementedError
    def load_objects(self) -> None: raise NotImplementedError

    def handle_events(self) -> None: raise NotImplementedError
    def handle_update(self) -> None: raise NotImplementedError
    def handle_render(self) -> None: raise NotImplementedError

    def run(self) -> None:
        self.load_assets()
        self.load_objects()
        while not self.events.quit:
            self.clock.update()
            self.events.update()

            self.handle_events()
            if isinstance(inputs.Mouse.Hovering, ui.Button) and self.events.mouse_pressed(inputs.Mouse.LeftClick):
                inputs.Mouse.Hovering.on_click()

            self.camera.update(self.clock.delta)
            
            self.handle_update()

            self.handle_render()
            self.renderer.flush()
            self.dev.render()

            for button in self.buttons.values():
                button.render(self.window.window)
                mouse_location = inputs.Mouse.get_location()
                if inputs.Mouse.Hovering == button:
                    if mouse_location[0] < button.location[0] or mouse_location[0] > button.location[0] + button.size[0]\
                    or mouse_location[1] < button.location[1] or mouse_location[1] > button.location[1] + button.size[1]:
                        inputs.Mouse.Hovering = None
                        button.on_unhover()
                elif inputs.Mouse.Hovering != button and\
                    mouse_location[0] >= button.location[0] and mouse_location[0] <= button.location[0] + button.size[0]\
                and mouse_location[1] >= button.location[1] and mouse_location[1] <= button.location[1] + button.size[1]:
                    inputs.Mouse.Hovering = button
                    button.on_hover()

            self.window.update()
            self.clock.rest()


from r3frame.globals import pg

class Button:
    def __init__(
            self, font_path:str, text: str="Button",
            size: list[int]=[64, 64], location: list[int|float]=[0, 0],
            color: list[int]=[0, 0, 0], text_color:list[int]=[255, 255, 255], text_size: int=18, padding: list[int]=[0, 0],
            border_size: list[int]=[5, 5], border_radius: list[int]=[0, 0, 0, 0], border_color: list[int]=[255, 255, 255],
        ) -> None:
        self.image = pg.Surface(size)

        self.size = size
        self.color = color
        self.location = location

        self.border_size = border_size
        self.border_radius = border_radius
        self.border_color = border_color
        
        self.text = text
        self.padding = padding
        self.font_path = font_path
        self.text_size: int = text_size
        self.text_color: list[int] = text_color
        self.font: pg.Font = pg.Font(font_path, self.text_size)

    def border(self) -> pg.Rect:
        return pg.Rect(
            [self.location[0] - self.border_size[0],
             self.location[1] - self.border_size[1],], 
            [self.size[0] + self.border_size[0],
             self.size[1] + self.border_size[1],]
        )
    
    def on_click(self) -> None: pass
    def on_hover(self) -> None: pass
    def on_unhover(self) -> None: pass

    def render(self, surface: pg.Surface) -> None:
        self.image.fill(self.color)
        self.image.blit(
        self.font.render(self.text, True, self.text_color, self.color), [self.padding[0], self.padding[1]])
        surface.blit(self.image, self.location)
        pg.draw.rect(surface, self.border_color, self.border(), width=self.border_size[0],
            border_top_left_radius=self.border_radius[0],
            border_top_right_radius=self.border_radius[1],
            border_bottom_left_radius=self.border_radius[2],
            border_bottom_right_radius=self.border_radius[3]
        )
